import copy
import logging
import time
from typing import List, Union
from datetime import timedelta, datetime

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.core import cache, config, utils
from spaceone.monitoring.model.alert_model import Alert
from spaceone.monitoring.model.project_alert_config_model import ProjectAlertConfig
from spaceone.monitoring.model.escalation_policy_model import EscalationPolicy
from spaceone.monitoring.manager.alert_manager import AlertManager
from spaceone.monitoring.manager.identity_manager import IdentityManager
from spaceone.monitoring.manager.webhook_manager import WebhookManager
from spaceone.monitoring.manager.maintenance_window_manager import MaintenanceWindowManager
from spaceone.monitoring.manager.project_alert_config_manager import ProjectAlertConfigManager
from spaceone.monitoring.manager.escalation_policy_manager import EscalationPolicyManager
from spaceone.monitoring.manager.notification_manager import NotificationManager
from spaceone.monitoring.manager.job_manager import JobManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class JobService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_mgr: JobManager = self.locator.get_manager('JobManager')

    @transaction(append_meta={'authorization.scope': 'SYSTEM'})
    def create_jobs_by_domain(self, params):
        """ Create jobs by domain

        Args:
            params (dict): {}

        Returns:
            None
        """

        for domain_id in self._list_domains_of_alerts():
            _LOGGER.debug(f'[create_jobs_by_domain] Push task (JobService.create): {domain_id}')
            self.job_mgr.push_task('monitoring_alert_job', 'JobService', 'create_job', {'domain_id': domain_id})

    @transaction(append_meta={'authorization.scope': 'SYSTEM'})
    @check_required(['domain_id'])
    def create_job(self, params):
        """ Create job

        Args:
            params (dict): {
                'domain_id': 'str'
            }

        Returns:
            None
        """

        domain_id = params['domain_id']

        if self.job_mgr.is_domain_job_running(domain_id):
            _LOGGER.debug(f'[create_job] Job is running. (domain_id = {domain_id})')
            return None

        job_vo = self.job_mgr.create_job(domain_id)
        try:
            alert_vos, total_count = self._list_open_alerts(domain_id)

            if total_count > 0:
                job_vo.update({'total_tasks': total_count, 'remained_tasks': total_count})

                for alert_vo in alert_vos:
                    _LOGGER.debug(f'[create_job] Push task (JobService.create_notification): {alert_vo.alert_id}')
                    self.job_mgr.push_task('monitoring_alert_notification_from_scheduler',
                                           'JobService',
                                           'create_alert_notification',
                                           {
                                               'job_id': job_vo.job_id,
                                               'alert_id': alert_vo.alert_id,
                                               'domain_id': alert_vo.domain_id
                                           })
            else:
                job_vo.delete()
        except Exception as e:
            self.job_mgr.change_error_status(job_vo, e)
            self.transaction.execute_rollback()

    @transaction(append_meta={'authorization.scope': 'SYSTEM'})
    @check_required(['alert_id', 'domain_id'])
    def create_assigned_notification(self, params):
        """ Create assigned notification

        Args:
            params (dict): {
                'alert_id': 'str',
                'domain_id': 'str',
                'user_id': 'str'
            }

        Returns:
            None
        """

        alert_id = params['alert_id']
        domain_id = params['domain_id']
        user_id = params['user_id']

        alert_mgr: AlertManager = self.locator.get_manager('AlertManager')

        alert_vo: Alert = alert_mgr.get_alert(alert_id, domain_id)

        title = f'[Assigned to me] {alert_vo.title}'

        notification_mgr: NotificationManager = self.locator.get_manager('NotificationManager')
        message = self._create_message(alert_vo, title, 'INFO', user_id=user_id)

        notification_mgr.create_notification(message)

    @transaction(append_meta={'authorization.scope': 'SYSTEM'})
    @check_required(['alert_id', 'domain_id'])
    def create_resolved_notification(self, params):
        """ Create resolved notification

        Args:
            params (dict): {
                'alert_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        alert_id = params['alert_id']
        domain_id = params['domain_id']

        alert_mgr: AlertManager = self.locator.get_manager('AlertManager')

        alert_vo: Alert = alert_mgr.get_alert(alert_id, domain_id)
        project_id = alert_vo.project_id
        escalation_policy_id = alert_vo.escalation_policy_id

        rules, finish_condition = self._get_escalation_policy_rules_and_finish_condition(escalation_policy_id,
                                                                                         domain_id)
        maintenance_window_state = self._get_project_maintenance_window_state(project_id, domain_id)

        if self._check_maintenance_window(project_id, alert_id, maintenance_window_state):
            title = f'[Resolved] {alert_vo.title}'

            for step in range(alert_vo.escalation_step):
                notification_level = rules[step]['notification_level']

                notification_mgr: NotificationManager = self.locator.get_manager('NotificationManager')
                message = self._create_message(alert_vo, title, 'SUCCESS', notification_level=notification_level)

                notification_mgr.create_notification(message)

    @transaction(append_meta={'authorization.scope': 'SYSTEM'})
    @check_required(['alert_id', 'domain_id'])
    def create_alert_notification(self, params):
        """ Create alert notification

        Args:
            params (dict): {
                'job_id': 'str',
                'alert_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        job_id = params.get('job_id')
        alert_id = params['alert_id']
        domain_id = params['domain_id']

        job_mgr: JobManager = self.locator.get_manager('JobManager')

        try:
            alert_mgr: AlertManager = self.locator.get_manager('AlertManager')

            alert_vo: Alert = alert_mgr.get_alert(alert_id, domain_id)
            project_id = alert_vo.project_id
            escalation_policy_id = alert_vo.escalation_policy_id

            alert_options = self._get_project_alert_options(project_id, domain_id)
            rules, finish_condition = self._get_escalation_policy_rules_and_finish_condition(escalation_policy_id,
                                                                                             domain_id)
            maintenance_window_state = self._get_project_maintenance_window_state(project_id, domain_id)

            # Check Notification Urgency and Finish Condition
            if not (self._check_notification_options(alert_vo.urgency, alert_id, alert_options)
                    and self._check_finish_condition(alert_vo.state, alert_id, finish_condition)
                    and self._check_maintenance_window(project_id, alert_id, maintenance_window_state)):
                alert_mgr.update_alert_by_vo({'escalation_ttl': 0}, alert_vo)

            else:
                # Escalate Alert
                is_notify, alert_vo = self._check_escalation_time_and_escalate_alert(alert_mgr, alert_vo, rules)
                if is_notify:
                    title = f'[Alerting] {alert_vo.title}'
                    current_rule = rules[alert_vo.escalation_step - 1]
                    notification_level = current_rule['notification_level']

                    notification_mgr: NotificationManager = self.locator.get_manager('NotificationManager')
                    message = self._create_message(alert_vo, title, 'ERROR', notification_level=notification_level,
                                                   has_callback=True, has_short_description=True)
                    notification_mgr.create_notification(message)

                    for project_id in alert_vo.project_dependencies:
                        dependent_project_message = copy.deepcopy(message)
                        dependent_project_message['resource_id'] = project_id
                        del dependent_project_message['message']['callbacks']
                        notification_mgr.create_notification(dependent_project_message)

            if job_id:
                job_vo = self.job_mgr.get_job(job_id, domain_id)
                job_mgr.decrease_remained_tasks(job_vo)
        except Exception as e:
            if job_id:
                job_vo = self.job_mgr.get_job(job_id, domain_id)
                job_mgr.change_error_status(job_vo, e)

            _LOGGER.error(f'[create_notification] Job Error: {e}', exc_info=True)
            self.transaction.execute_rollback()

    def _list_domains_of_alerts(self):
        alert_mgr: AlertManager = self.locator.get_manager('AlertManager')
        query = {
            'distinct': 'domain_id',
            'filter': [
                {
                    'k': 'state',
                    'v': ['TRIGGERED', 'ACKNOWLEDGED'],
                    'o': 'in'
                },
                {
                    'k': 'escalation_ttl',
                    'v': 0,
                    'o': 'gt'
                }
            ]
        }

        response = alert_mgr.stat_alerts(query)
        return response.get('results', [])

    def _list_open_alerts(self, domain_id) -> List[Alert]:
        alert_mgr: AlertManager = self.locator.get_manager('AlertManager')
        project_alert_config_mgr: ProjectAlertConfigManager = self.locator.get_manager('ProjectAlertConfigManager')

        project_ids = []
        project_alert_config_vos, total_count = project_alert_config_mgr.list_project_alert_configs({
            'filter': [
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                }
            ],
            'only': ['project_id']
        })

        for project_alert_config_vo in project_alert_config_vos:
            project_ids.append(project_alert_config_vo.project_id)

        query = {
            'filter': [
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                },
                {
                    'k': 'project_id',
                    'v': project_ids,
                    'o': 'in'
                },
                {
                    'k': 'state',
                    'v': ['TRIGGERED', 'ACKNOWLEDGED'],
                    'o': 'in'
                },
                {
                    'k': 'escalation_ttl',
                    'v': 0,
                    'o': 'gt'
                }
            ]
        }

        return alert_mgr.list_alerts(query)

    @cache.cacheable(key='project-alert-options:{domain_id}:{project_id}', expire=300)
    def _get_project_alert_options(self, project_id, domain_id):
        project_alert_config_mgr: ProjectAlertConfigManager = self.locator.get_manager('ProjectAlertConfigManager')
        project_alert_config_vo: ProjectAlertConfig = project_alert_config_mgr.get_project_alert_config(project_id,
                                                                                                        domain_id)

        return dict(project_alert_config_vo.options.to_dict())

    @cache.cacheable(key='escalation-policy-condition:{domain_id}:{escalation_policy_id}', expire=300)
    def _get_escalation_policy_rules_and_finish_condition(self, escalation_policy_id, domain_id):
        escalation_policy_mgr: EscalationPolicyManager = self.locator.get_manager('EscalationPolicyManager')
        escalation_policy_vo: EscalationPolicy = escalation_policy_mgr.get_escalation_policy(escalation_policy_id,
                                                                                             domain_id)
        rules = []
        for rule in escalation_policy_vo.rules:
            rules.append(dict(rule.to_dict()))

        return rules, escalation_policy_vo.finish_condition

    @cache.cacheable(key='maintenance-window-state:{domain_id}:{project_id}', expire=60)
    def _get_project_maintenance_window_state(self, project_id, domain_id):
        query = {
            'filter': [
                {
                    'k': 'projects',
                    'v': project_id,
                    'o': 'eq'
                },
                {
                    'k': 'state',
                    'v': 'OPEN',
                    'o': 'eq'
                },
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                }
            ],
            'count_only': True
        }

        maintenance_window_mgr: MaintenanceWindowManager = self.locator.get_manager('MaintenanceWindowManager')
        maintenance_window_vos, total_count = maintenance_window_mgr.list_maintenance_windows(query)

        if total_count > 0:
            return 'OPEN'
        else:
            return 'CLOSED'

    @staticmethod
    def _get_current_escalation_rule(alert_vo: Alert, rules):
        return rules[alert_vo.escalation_step - 1]

    @staticmethod
    def _check_notification_options(alert_urgency, alert_id, alert_options):
        if alert_options['notification_urgency'] == 'HIGH_ONLY' and alert_urgency == 'LOW':
            _LOGGER.debug(f'[_check_notification_options] Stop notifications. '
                          f'(notification_urgency = HIGH_ONLY, alert_urgency = LOW, alert_id = {alert_id})')
            return False
        else:
            return True

    @staticmethod
    def _check_finish_condition(alert_state, alert_id, finish_condition):
        if finish_condition == 'ACKNOWLEDGED' and alert_state == 'ACKNOWLEDGED':
            _LOGGER.debug(f'[_check_finish_condition] Stop notifications. '
                          f'(finish_condition = ACKNOWLEDGED, alert_state = ACKNOWLEDGED, '
                          f'alert_id = {alert_id})')
            return False
        else:
            return True

    @staticmethod
    def _check_maintenance_window(project_id, alert_id, maintenance_window_state):
        if maintenance_window_state == 'OPEN':
            _LOGGER.debug(f'[_check_maintenance_window] Stop notifications. '
                          f'(project_id = {project_id}, maintenance_window_state = OPEN, alert_id = {alert_id})')
            return False
        else:
            return True

    @staticmethod
    def _check_escalation_time_and_escalate_alert(alert_mgr: AlertManager, alert_vo: Alert, rules):
        current_step = alert_vo.escalation_step
        escalation_ttl = alert_vo.escalation_ttl
        current_rule = rules[current_step - 1]
        escalate_minutes = current_rule.get('escalate_minutes', 0)
        escalated_at: Union[datetime, None] = alert_vo.escalated_at

        # First triggered alert
        if escalated_at is None:
            escalated_alert_vo = alert_mgr.update_alert_by_vo({'escalated_at': datetime.utcnow()}, alert_vo)
            return True, escalated_alert_vo
        else:
            now = datetime.utcnow()

            # now > escalated_at + escalate_minutes
            if now > (escalated_at + timedelta(minutes=escalate_minutes)):
                # When the current step is the maximum
                if len(rules) == current_step:
                    if escalation_ttl == 1:
                        _LOGGER.debug(f'[_check_escalation_time_and_escalate_alert] Max escalation step. '
                                      f'(alert_id = {alert_vo.alert_id})')

                        escalated_alert_vo = alert_mgr.update_alert_by_vo({
                            'escalated_at': datetime.utcnow(),
                            'escalation_ttl': escalation_ttl - 1
                        }, alert_vo)

                        return False, escalated_alert_vo
                    else:
                        _LOGGER.debug(f'[_check_escalation_time_and_escalate_alert] Repeat again from the first step. '
                                      f'(alert_id = {alert_vo.alert_id})')

                        escalated_alert_vo = alert_mgr.update_alert_by_vo({
                            'escalated_at': datetime.utcnow(),
                            'escalation_step': 1,
                            'escalation_ttl': escalation_ttl - 1
                        }, alert_vo)

                        return True, escalated_alert_vo
                else:
                    _LOGGER.debug(f'[_check_escalation_time_and_escalate_alert] Escalate from {current_step} '
                                  f'to {current_step + 1} steps. (alert_id = {alert_vo.alert_id})')

                    escalated_alert_vo = alert_mgr.update_alert_by_vo({
                        'escalated_at': datetime.utcnow(),
                        'escalation_step': current_step + 1
                    }, alert_vo)

                    return True, escalated_alert_vo
            else:
                return False, alert_vo

    def _create_message(self, alert_vo: Alert, title: str, notification_type: str, notification_level='ALL',
                        has_callback=False, has_short_description=False, user_id=None):

        domain_id = alert_vo.domain_id
        project_name = self._get_project_name(alert_vo.project_id, domain_id)

        tags = [
            {
                'key': 'Alert Number',
                'value': f'#{alert_vo.alert_number}',
                'options': {
                    'short': True
                }
            },
            {
                'key': 'State',
                'value': alert_vo.state,
                'options': {
                    'short': True
                }
            },
            {
                'key': 'Urgency',
                'value': alert_vo.urgency,
                'options': {
                    'short': True
                }
            },
            {
                'key': 'Triggered by',
                'value': self._get_triggered_by_name(alert_vo.triggered_by, domain_id),
                'options': {
                    'short': True
                }
            },
            {
                'key': 'Project',
                'value': project_name
            }
        ]

        resource = alert_vo.resource or {}

        if 'name' in resource:
            tags.append({
                'key': 'Resource',
                'value': resource['name']
            })

        if alert_vo.assignee:
            tags.append({
                'key': 'Assigned to',
                'value': self._get_user_name(alert_vo.assignee, domain_id)
            })

        if alert_vo.status_message and alert_vo.status_message != '':
            tags.append({
                'key': 'Status Message',
                'value': alert_vo.status_message
            })

        if alert_vo.account:
            tags.append({
                'key': 'Account',
                'value': alert_vo.account
            })

        callbacks = []

        if has_callback:
            access_key = self._generate_access_key()
            callback_url = self._make_callback_url(alert_vo.alert_id, domain_id, access_key)
            callbacks.append(
                {
                    'label': 'Acknowledge Alerts',
                    'url': callback_url
                }
            )

        if user_id:
            resource_type = 'identity.User'
            resource_id = user_id
        else:
            resource_type = 'identity.Project'
            resource_id = alert_vo.project_id

        message = {
            'title': title,
            'tags': tags,
            'callbacks': callbacks,
            'occurred_at': utils.datetime_to_iso8601(alert_vo.created_at)
        }

        if alert_vo.description and alert_vo.description != '':
            message['description'] = alert_vo.description

        alert_link = self._make_alert_link(alert_vo.alert_id, alert_vo.domain_id)

        if alert_link:
            message['link'] = alert_link

        if alert_vo.image_url:
            message['image_url'] = alert_vo.image_url

        if has_short_description:
            # TODO: Need to change multiple language
            message['short_description'] = f'경고! 장애 발생! {project_name.replace(" >", "의")} 프로젝트에 장애가 발생했습니다.'

        return {
            'resource_type': resource_type,
            'resource_id': resource_id,
            "notification_type": notification_type,
            'topic': 'monitoring.Alert',
            'message': message,
            'notification_level': notification_level,
            'domain_id': domain_id
        }

    @cache.cacheable(key='project-name:{domain_id}:{project_id}', expire=300)
    def _get_project_name(self, project_id, domain_id):
        try:
            identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
            project_info = identity_mgr.get_project(project_id, domain_id)
            return f'{project_info["project_group_info"]["name"]} > {project_info["name"]}'
        except Exception as e:
            _LOGGER.error(f'[_get_project_name] Failed to get project: {e}', exc_info=True)

        return project_id

    @cache.cacheable(key='triggered-by-name:{domain_id}:{triggered_by}', expire=300)
    def _get_triggered_by_name(self, triggered_by, domain_id):
        if triggered_by.startswith('webhook-'):
            try:
                webhook_mgr: WebhookManager = self.locator.get_manager('WebhookManager')
                webhook_info = webhook_mgr.get_webhook(triggered_by, domain_id)

                return webhook_info['name']
            except Exception as e:
                _LOGGER.error(f'[_get_triggered_by_name] Failed to get webhook: {e}', exc_info=True)
        else:
            return self._get_user_name(triggered_by, domain_id)

        return triggered_by

    @cache.cacheable(key='user-name:{domain_id}:{user_id}', expire=300)
    def _get_user_name(self, user_id, domain_id):
        try:
            identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
            user_info = identity_mgr.get_user(user_id, domain_id)

            if len(user_info.get('name', '').strip()) == 0:
                return user_info['user_id']
            else:
                return f'{user_info["user_id"]} ({user_info["name"]})'
        except Exception as e:
            _LOGGER.error(f'[_get_user_name] Failed to get user: {e}', exc_info=True)

        return user_id

    @staticmethod
    def _generate_access_key():
        return utils.random_string(16)

    def _make_callback_url(self, alert_id, domain_id, access_key):
        def _rollback(alert_id, access_key):
            _LOGGER.info(f'[_make_callback_url._rollback] '
                         f'Delete cache : {alert_id} '
                         f'({access_key})')
            cache.delete(f'alert-notification-callback:{alert_id}:{access_key}')

        cache.set(f'alert-notification-callback:{alert_id}:{access_key}', domain_id, expire=600)
        self.transaction.add_rollback(_rollback, alert_id, access_key)

        webhook_domain = config.get_global('WEBHOOK_DOMAIN')
        return f'{webhook_domain}/monitoring/v1/alert/{alert_id}/{access_key}/ACKNOWLEDGED'

    def _make_alert_link(self, alert_id, domain_id):
        console_domain: str = config.get_global('CONSOLE_DOMAIN')

        if console_domain.strip() != '':
            domain_name = self._get_domain_name(domain_id)
            console_domain = console_domain.format(domain_name=domain_name)

            return f'{console_domain}/alert-manager/alert/{alert_id}'

    @cache.cacheable(key='domain-name:{domain_id}', expire=3600)
    def _get_domain_name(self, domain_id):
        try:
            identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
            domain_info = identity_mgr.get_domain(domain_id)
            return domain_info['name']
        except Exception as e:
            _LOGGER.error(f'[_get_domain_name] Failed to get domain: {e}', exc_info=True)

        return domain_id



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class Responder(EmbeddedDocument):
    resource_type = StringField(max_length=255)
    resource_id = StringField(max_length=255)


class AlertResource(EmbeddedDocument):
    resource_id = StringField(default=None, null=True)
    resource_type = StringField(default=None, null=True)
    name = StringField(default=None, null=True)


class AlertNumber(MongoModel):
    next = IntField(default=1)
    domain_id = StringField(max_length=40)


class Alert(MongoModel):
    alert_number = IntField(required=True)
    alert_number_str = StringField()
    alert_id = StringField(max_length=40, generate_id='alert', unique=True)
    title = StringField()
    state = StringField(max_length=20, default='TRIGGERED', choices=('TRIGGERED', 'ACKNOWLEDGED', 'RESOLVED', 'ERROR'))
    status_message = StringField(default=None, null=True)
    description = StringField(default=None, null=True)
    assignee = StringField(default=None, null=True)
    urgency = StringField(max_length=20, default='HIGH', choices=('HIGH', 'LOW'))
    severity = StringField(max_length=20, default='NONE', choices=('CRITICAL', 'ERROR', 'WARNING', 'INFO',
                                                                   'NOT_AVAILABLE', 'NONE'))
    rule = StringField(default=None, null=True)
    image_url = StringField(default=None, null=True)
    resource = EmbeddedDocumentField(AlertResource, default=None, null=True)
    provider = StringField(default=None, null=True)
    account = StringField(default=None, null=True)
    additional_info = DictField()
    is_snoozed = BooleanField(default=False)
    snoozed_end_time = DateTimeField(default=None, null=True)
    escalation_step = IntField(default=1)
    escalation_ttl = IntField(default=0)
    responders = ListField(EmbeddedDocumentField(Responder))
    project_dependencies = ListField(StringField(max_length=40))
    triggered_by = StringField(default=None, null=True)
    webhook_id = StringField(max_length=40, default=None, null=True)
    escalation_policy_id = StringField(max_length=40)
    project_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)
    acknowledged_at = DateTimeField(default=None, null=True)
    resolved_at = DateTimeField(default=None, null=True)
    escalated_at = DateTimeField(default=None, null=True)

    meta = {
        'updatable_fields': [
            'title',
            'state',
            'status_message',
            'description',
            'assignee',
            'urgency',
            'is_snoozed',
            'snoozed_end_time',
            'escalation_step',
            'escalation_ttl',
            'responders',
            'project_dependencies',
            'escalation_policy_id',
            'project_id',
            'acknowledged_at',
            'resolved_at',
            'escalated_at'
        ],
        'minimal_fields': [
            'alert_number',
            'alert_id',
            'title',
            'state',
            'status_message',
            'assignee',
            'urgency',
            'escalation_step',
            'escalation_ttl',
            'project_id'
        ],
        'change_query_keys': {
            'resource_id': 'resource.resource_id',
            'user_projects': 'project_id'
        },
        'ordering': [
            '-created_at'
        ],
        'indexes': [
            'alert_number',
            # 'alert_id',
            'state',
            'assignee',
            'urgency',
            'severity',
            'is_snoozed',
            'snoozed_end_time',
            'resource.resource_id',
            'resource.resource_type',
            'resource.name',
            'provider',
            'account',
            'escalation_step',
            'responders.resource_type',
            'responders.resource_id',
            'project_dependencies',
            'triggered_by',
            'webhook_id',
            'escalation_policy_id',
            'project_id',
            'domain_id',
            'created_at',
            'acknowledged_at',
            'resolved_at',
            'escalated_at',
            {
                "fields": ['domain_id', 'state', 'is_snoozed', 'escalation_step', 'escalation_ttl',
                           'escalation_policy_id', 'escalated_at'],
                "name": "COMPOUND_INDEX_FOR_ESCALATION"
            },
        ]
    }



import logging

from spaceone.core.service import *
from spaceone.core import utils

from spaceone.monitoring.error import *
from spaceone.monitoring.model.webhook_model import Webhook
from spaceone.monitoring.manager.project_alert_config_manager import ProjectAlertConfigManager
from spaceone.monitoring.manager.repository_manager import RepositoryManager
from spaceone.monitoring.manager.webhook_plugin_manager import WebhookPluginManager
from spaceone.monitoring.manager.webhook_manager import WebhookManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class WebhookService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.webhook_mgr: WebhookManager = self.locator.get_manager('WebhookManager')
        self.webhook_plugin_mgr: WebhookPluginManager = self.locator.get_manager('WebhookPluginManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['name', 'plugin_info', 'project_id', 'domain_id'])
    def create(self, params):
        """Create webhook

        Args:
            params (dict): {
                'name': 'str',
                'plugin_info': 'dict',
                'project_id': 'str'
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            webhook_vo (object)
        """

        domain_id = params['domain_id']
        project_id = params['project_id']

        project_alert_config_mgr: ProjectAlertConfigManager = self.locator.get_manager('ProjectAlertConfigManager')

        project_alert_config_mgr.get_project_alert_config(project_id, domain_id)

        self._check_plugin_info(params['plugin_info'])
        plugin_info = self._get_plugin(params['plugin_info'], domain_id)
        params['capability'] = plugin_info.get('capability', {})

        _LOGGER.debug(f'[create] Init Plugin: {params["plugin_info"]}')
        endpoint, updated_version = self.webhook_plugin_mgr.get_webhook_plugin_endpoint(params['plugin_info'], domain_id)
        if updated_version:
            params['plugin_info']['version'] = updated_version

        options = params['plugin_info'].get('options', {})
        plugin_metadata = self._init_plugin(endpoint, options)
        params['plugin_info']['metadata'] = plugin_metadata

        webhook_vo: Webhook = self.webhook_mgr.create_webhook(params)

        access_key = self._generate_access_key()
        webhook_url = self._make_webhook_url(webhook_vo.webhook_id, access_key)

        return self.webhook_mgr.update_webhook_by_vo({
            'access_key': access_key,
            'webhook_url': webhook_url
        }, webhook_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def update(self, params):
        """Update webhook

        Args:
            params (dict): {
                'webhook_id': 'str',
                'name': 'dict',
                'tags': 'dict'
                'domain_id': 'str'
            }

        Returns:
            webhook_vo (object)
        """

        webhook_id = params['webhook_id']
        domain_id = params['domain_id']
        webhook_vo = self.webhook_mgr.get_webhook(webhook_id, domain_id)

        return self.webhook_mgr.update_webhook_by_vo(params, webhook_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def enable(self, params):
        """ Enable webhook

        Args:
            params (dict): {
                'webhook_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            webhook_vo (object)
        """

        webhook_id = params['webhook_id']
        domain_id = params['domain_id']
        webhook_vo = self.webhook_mgr.get_webhook(webhook_id, domain_id)

        return self.webhook_mgr.update_webhook_by_vo({'state': 'ENABLED'}, webhook_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def disable(self, params):
        """ Disable webhook

        Args:
            params (dict): {
                'webhook_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            webhook_vo (object)
        """

        webhook_id = params['webhook_id']
        domain_id = params['domain_id']
        webhook_vo = self.webhook_mgr.get_webhook(webhook_id, domain_id)

        return self.webhook_mgr.update_webhook_by_vo({'state': 'DISABLED'}, webhook_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def delete(self, params):
        """Delete webhook

        Args:
            params (dict): {
                'webhook_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.webhook_mgr.delete_webhook(params['webhook_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def verify_plugin(self, params):
        """ Verify webhook plugin

        Args:
            params (dict): {
                'webhook_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            webhook_vo (object)
        """

        webhook_id = params['webhook_id']
        domain_id = params['domain_id']
        webhook_vo = self.webhook_mgr.get_webhook(webhook_id, domain_id)

        endpoint = self.webhook_plugin_mgr.get_webhook_plugin_endpoint_by_vo(webhook_vo)

        self._verify_plugin(endpoint, webhook_vo.plugin_info.options)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def update_plugin(self, params):
        """Update webhook plugin

        Args:
            params (dict): {
                'webhook_id': 'str',
                'version': 'str',
                'options': 'dict',
                'upgrade_mode': 'str',
                'domain_id': 'str'
            }

        Returns:
            webhook_vo (object)
        """

        webhook_id = params['webhook_id']
        domain_id = params['domain_id']
        options = params.get('options')
        version = params.get('version')
        upgrade_mode = params.get('upgrade_mode')

        webhook_vo = self.webhook_mgr.get_webhook(webhook_id, domain_id)
        plugin_info = webhook_vo.plugin_info.to_dict()

        if version:
            plugin_info['version'] = version

        if options:
            plugin_info['options'] = options

        if upgrade_mode:
            plugin_info['upgrade_mode'] = upgrade_mode

        endpoint, updated_version = self.webhook_plugin_mgr.get_webhook_plugin_endpoint(plugin_info, domain_id)
        if updated_version:
            plugin_info['version'] = updated_version

        plugin_metadata = self._init_plugin(endpoint, plugin_info.get('options', {}))
        plugin_info['metadata'] = plugin_metadata

        params = {
            'plugin_info': plugin_info
        }

        _LOGGER.debug(f'[update_plugin] {plugin_info}')

        return self.webhook_mgr.update_webhook_by_vo(params, webhook_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['webhook_id', 'domain_id'])
    def get(self, params):
        """ Get webhook

        Args:
            params (dict): {
                'webhook_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            webhook_vo (object)
        """

        return self.webhook_mgr.get_webhook(params['webhook_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['webhook_id', 'name', 'state', 'access_key', 'project_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['webhook_id', 'name'])
    def list(self, params):
        """ List webhooks

        Args:
            params (dict): {
                'webhook_id': 'str',
                'name': 'str',
                'state': 'str',
                'project_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', // from meta
            }

        Returns:
            webhook_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.webhook_mgr.list_webhooks(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['webhook_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.webhook_mgr.stat_webhooks(query)

    @staticmethod
    def _generate_access_key():
        return utils.random_string(16)

    @staticmethod
    def _make_webhook_url(webhook_id, access_key):
        return f'/monitoring/v1/webhook/{webhook_id}/{access_key}/events'

    @staticmethod
    def _check_plugin_info(plugin_info_params):
        if 'plugin_id' not in plugin_info_params:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.plugin_id')

    def _get_plugin(self, plugin_info, domain_id):
        plugin_id = plugin_info['plugin_id']

        repo_mgr: RepositoryManager = self.locator.get_manager('RepositoryManager')
        plugin_info = repo_mgr.get_plugin(plugin_id, domain_id)

        return plugin_info

    def _init_plugin(self, endpoint, options):
        self.webhook_plugin_mgr.initialize(endpoint)
        return self.webhook_plugin_mgr.init_plugin(options)

    def _verify_plugin(self, endpoint, options):
        self.webhook_plugin_mgr.initialize(endpoint)
        self.webhook_plugin_mgr.verify_plugin(options)



import logging

from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.monitoring.error import *

_LOGGER = logging.getLogger(__name__)


class InventoryManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inventory_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='inventory')

    def get_server(self, server_id, domain_id):
        return self.inventory_connector.dispatch('Server.get', {'server_id': server_id, 'domain_id': domain_id})

    def list_servers(self, query, domain_id):
        return self.inventory_connector.dispatch('Server.list', {'query': query, 'domain_id': domain_id})

    def get_cloud_service(self, cloud_service_id, domain_id):
        return self.inventory_connector.dispatch('CloudService.get', {'cloud_service_id': cloud_service_id,
                                                                      'domain_id': domain_id})

    def list_cloud_services(self, query, domain_id):
        return self.inventory_connector.dispatch('CloudService.list', {'query': query, 'domain_id': domain_id})

    def get_resource(self, resource_type, resource_id, domain_id):
        return self.get_cloud_service(resource_id, domain_id)

    def list_resources(self, resources, required_keys, domain_id):
        query = self._make_query(resources, required_keys)
        response = self.list_cloud_services(query, domain_id)
        return response.get('results', [])

    @staticmethod
    def _make_query(resources, required_keys):
        only_keys = list(set(['cloud_service_id', 'collection_info.secret_id', 'region_code'] + required_keys))
        only_keys.sort()

        return {
            'filter': [{
                'k': 'cloud_service_id',
                'v': resources,
                'o': 'in'
            }],
            'only': only_keys
        }



import logging

from spaceone.core.service import *
from spaceone.core import cache
from spaceone.core import config
from spaceone.core import utils
from spaceone.monitoring.error import *
from spaceone.monitoring.manager.repository_manager import RepositoryManager
from spaceone.monitoring.manager.secret_manager import SecretManager
from spaceone.monitoring.manager.data_source_plugin_manager import DataSourcePluginManager
from spaceone.monitoring.manager.data_source_manager import DataSourceManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class DataSourceService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_mgr: DataSourceManager = self.locator.get_manager('DataSourceManager')
        self.ds_plugin_mgr: DataSourcePluginManager = self.locator.get_manager('DataSourcePluginManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['name', 'plugin_info', 'domain_id'])
    def register(self, params):
        """Register data source

        Args:
            params (dict): {
                'name': 'str',
                'plugin_info': 'dict',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        domain_id = params['domain_id']

        self._check_plugin_info(params['plugin_info'])
        plugin_info = self._get_plugin(params['plugin_info'], domain_id)
        params['capability'] = plugin_info.get('capability', {})
        params['provider'] = plugin_info.get('provider')

        self._check_plugin_capability(params['capability'])
        params['monitoring_type'] = params['capability']['monitoring_type']

        # Update metadata
        endpoint, updated_version = self.ds_plugin_mgr.get_data_source_plugin_endpoint(params['plugin_info'], domain_id)
        if updated_version:
            params['plugin_info']['version'] = updated_version

        options = params['plugin_info'].get('options', {})
        plugin_metadata = self._init_plugin(endpoint, options, params['monitoring_type'])
        params['plugin_info']['metadata'] = plugin_metadata

        return self.data_source_mgr.register_data_source(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def update(self, params):
        """Update data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'dict',
                'tags': 'dict'
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """
        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        return self.data_source_mgr.update_data_source_by_vo(params, data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def enable(self, params):
        """ Enable data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        return self.data_source_mgr.update_data_source_by_vo({'state': 'ENABLED'},
                                                             data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def disable(self, params):
        """ Disable data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        return self.data_source_mgr.update_data_source_by_vo({'state': 'DISABLED'},
                                                             data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def deregister(self, params):
        """Deregister data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.data_source_mgr.deregister_data_source(params['data_source_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def verify_plugin(self, params):
        """ Verify data source plugin

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        endpoint = self.ds_plugin_mgr.get_data_source_plugin_endpoint_by_vo(data_source_vo)

        self._verify_plugin(endpoint, data_source_vo.plugin_info.to_dict(), data_source_vo.capability, domain_id)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def update_plugin(self, params):
        """Update data source plugin

        Args:
            params (dict): {
                'data_source_id': 'str',
                'version': 'str',
                'options': 'dict',
                'upgrade_mode': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """
        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        version = params.get('version')
        options = params.get('options')
        upgrade_mode = params.get('upgrade_mode')

        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        plugin_info = data_source_vo.plugin_info.to_dict()

        if version:
            plugin_info['version'] = version

        if options:
            plugin_info['options'] = options

        if upgrade_mode:
            plugin_info['upgrade_mode'] = upgrade_mode

        endpoint, updated_version = self.ds_plugin_mgr.get_data_source_plugin_endpoint(plugin_info, domain_id)
        if updated_version:
            plugin_info['version'] = updated_version

        plugin_metadata = self._init_plugin(endpoint, plugin_info.get('options', {}), data_source_vo.monitoring_type)
        plugin_info['metadata'] = plugin_metadata

        params = {
            'plugin_info': plugin_info
        }

        _LOGGER.debug(f'[update_plugin] {plugin_info}')

        return self.data_source_mgr.update_data_source_by_vo(params, data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def get(self, params):
        """ Get data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            data_source_vo (object)
        """
        domain_id = params['domain_id']
        self._initialize_data_sources(domain_id)

        return self.data_source_mgr.get_data_source(params['data_source_id'], domain_id, params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['data_source_id', 'name', 'state', 'monitoring_type', 'provider', 'domain_id'])
    @append_keyword_filter(['data_source_id', 'name', 'provider'])
    def list(self, params):
        """ List data sources

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'state': 'str',
                'monitoring_type': 'str',
                'provider': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            data_source_vos (object)
            total_count
        """
        domain_id = params['domain_id']
        query = params.get('query', {})
        self._initialize_data_sources(domain_id)

        return self.data_source_mgr.list_data_sources(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['data_source_id', 'name', 'provider'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.data_source_mgr.stat_data_sources(query)

    @staticmethod
    def _check_plugin_capability(capability):
        if 'monitoring_type' not in capability:
            raise ERROR_WRONG_PLUGIN_SETTINGS(key='capability.monitoring_type')
        else:
            if capability['monitoring_type'] not in ['METRIC', 'LOG']:
                raise ERROR_WRONG_PLUGIN_SETTINGS(key='capability.monitoring_type')

        if 'supported_schema' not in capability:
            raise ERROR_WRONG_PLUGIN_SETTINGS(key='capability.supported_schema')

    @staticmethod
    def _check_plugin_info(plugin_info_params):
        if 'plugin_id' not in plugin_info_params:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.plugin_id')

        if plugin_info_params.get('upgrade_mode', 'AUTO') == 'MANUAL' and 'version' not in plugin_info_params:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.version')

        secret_id = plugin_info_params.get('secret_id')
        provider = plugin_info_params.get('provider')

        if secret_id is None and provider is None:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.[secret_id | provider]')

    def _get_plugin(self, plugin_info, domain_id):
        plugin_id = plugin_info['plugin_id']

        repo_mgr: RepositoryManager = self.locator.get_manager('RepositoryManager')
        return repo_mgr.get_plugin(plugin_id, domain_id)

    def _init_plugin(self, endpoint, options, monitoring_type):
        self.ds_plugin_mgr.initialize(endpoint)
        return self.ds_plugin_mgr.init_plugin(options, monitoring_type)

    def _verify_plugin(self, endpoint, plugin_info, capability, domain_id):
        options = plugin_info.get('options', {})
        secret_mgr: SecretManager = self.locator.get_manager('SecretManager')
        secret_data, schema = secret_mgr.get_secret_data_from_plugin(plugin_info, capability, domain_id)

        ds_plugin_mgr: DataSourcePluginManager = self.locator.get_manager('DataSourcePluginManager')
        ds_plugin_mgr.initialize(endpoint)
        ds_plugin_mgr.verify_plugin(options, secret_data, schema)

    @cache.cacheable(key='init-data-source:{domain_id}', expire=300)
    def _initialize_data_sources(self, domain_id):
        _LOGGER.debug(f'[_initialize_data_source] domain_id: {domain_id}')

        query = {'filter': [{'k': 'domain_id', 'v': domain_id, 'o': 'eq'}]}
        data_source_vos, total_count = self.data_source_mgr.list_data_sources(query)

        installed_data_sources_ids = [data_source_vo.plugin_info.plugin_id for data_source_vo in data_source_vos]
        _LOGGER.debug(f'[_initialize_data_source] Installed Plugins : {installed_data_sources_ids}')

        global_conf = config.get_global()
        for _data_source in global_conf.get('INSTALLED_DATA_SOURCE_PLUGINS', []):
            if _data_source['plugin_info']['plugin_id'] not in installed_data_sources_ids:
                try:
                    _LOGGER.debug(
                        f'[_initialize_data_source] Create init data source: {_data_source["plugin_info"]["plugin_id"]}')
                    _data_source['domain_id'] = domain_id
                    self.register(_data_source)
                except Exception as e:
                    _LOGGER.error(f'[_initialize_data_source] {e}')

        return True



import logging

from spaceone.core.manager import BaseManager
from spaceone.monitoring.error import *
from spaceone.monitoring.manager.plugin_manager import PluginManager
from spaceone.monitoring.connector.datasource_plugin_connector import DataSourcePluginConnector
from spaceone.monitoring.model.data_source_model import DataSource
from spaceone.monitoring.model.plugin_metadata_model import MetricPluginMetadataModel, LogPluginMetadataModel

_LOGGER = logging.getLogger(__name__)


class DataSourcePluginManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dsp_connector: DataSourcePluginConnector = self.locator.get_connector('DataSourcePluginConnector')

    def initialize(self, endpoint):
        _LOGGER.debug(f'[initialize] data source plugin endpoint: {endpoint}')
        self.dsp_connector.initialize(endpoint)

    def init_plugin(self, options, monitoring_type):
        plugin_info = self.dsp_connector.init(options)

        _LOGGER.debug(f'[plugin_info] {plugin_info}')
        plugin_metadata = plugin_info.get('metadata', {})

        self._validate_plugin_metadata(plugin_metadata, monitoring_type)

        return plugin_metadata

    def verify_plugin(self, options, secret_data, schema):
        self.dsp_connector.verify(options, secret_data, schema)

    def list_metrics(self, schema, options, secret_data, query):
        return self.dsp_connector.list_metrics(schema, options, secret_data, query)

    def get_metric_data(self, params):
        return self.dsp_connector.get_metric_data(params)

    def list_logs(self, schema, options, secret_data, query, keyword, start, end, sort, limit):
        results = []

        for result in self.dsp_connector.list_logs(schema, options, secret_data, query, keyword, start, end, sort, limit):
            results.extend(result.get('results', []))

        return {
            'results': results
        }

    @staticmethod
    def _validate_plugin_metadata(plugin_metadata, monitoring_type):
        try:
            if monitoring_type == 'METRIC':
                MetricPluginMetadataModel(plugin_metadata).validate()
            else:
                LogPluginMetadataModel(plugin_metadata).validate()

        except Exception as e:
            raise ERROR_INVALID_PLUGIN_OPTIONS(reason=str(e))

    @staticmethod
    def _process_stream(response_stream):
        for response in response_stream:
            yield response

    def get_data_source_plugin_endpoint_by_vo(self, data_source_vo: DataSource):
        plugin_info = data_source_vo.plugin_info.to_dict()
        endpoint, updated_version = self.get_data_source_plugin_endpoint(plugin_info, data_source_vo.domain_id)

        if updated_version:
            _LOGGER.debug(f'[get_data_source_plugin_endpoint_by_vo] upgrade plugin version: {plugin_info["version"]} -> {updated_version}')
            self.upgrade_data_source_plugin_version(data_source_vo, endpoint, updated_version)

        return endpoint

    def get_data_source_plugin_endpoint(self, plugin_info, domain_id):
        plugin_mgr: PluginManager = self.locator.get_manager('PluginManager')
        return plugin_mgr.get_plugin_endpoint(plugin_info, domain_id)

    def upgrade_data_source_plugin_version(self, data_source_vo: DataSource, endpoint, updated_version):
        plugin_info = data_source_vo.plugin_info.to_dict()
        self.initialize(endpoint)
        plugin_metadata = self.init_plugin(plugin_info.get('options', {}), data_source_vo.monitoring_type)
        plugin_info['version'] = updated_version
        plugin_info['metadata'] = plugin_metadata
        data_source_vo.update({'plugin_info': plugin_info})



PROTO = {
    'spaceone.monitoring.interface.grpc.v1.data_source': ['DataSource'],
    'spaceone.monitoring.interface.grpc.v1.metric': ['Metric'],
    'spaceone.monitoring.interface.grpc.v1.log': ['Log'],
    'spaceone.monitoring.interface.grpc.v1.project_alert_config': ['ProjectAlertConfig'],
    'spaceone.monitoring.interface.grpc.v1.escalation_policy': ['EscalationPolicy'],
    'spaceone.monitoring.interface.grpc.v1.event_rule': ['EventRule'],
    'spaceone.monitoring.interface.grpc.v1.webhook': ['Webhook'],
    'spaceone.monitoring.interface.grpc.v1.maintenance_window': ['MaintenanceWindow'],
    'spaceone.monitoring.interface.grpc.v1.alert': ['Alert'],
    'spaceone.monitoring.interface.grpc.v1.note': ['Note'],
    'spaceone.monitoring.interface.grpc.v1.event': ['Event'],
}



from spaceone.api.monitoring.v1 import alert_pb2, alert_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Alert(BaseAPI, alert_pb2_grpc.AlertServicer):

    pb2 = alert_pb2
    pb2_grpc = alert_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.update(params))

    def update_state(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.update_state(params))

    def merge(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.merge(params))

    def snooze(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.snooze(params))

    def add_responder(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.add_responder(params))

    def remove_responder(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.remove_responder(params))

    def add_project_dependency(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.add_project_dependency(params))

    def remove_project_dependency(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.remove_project_dependency(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            alert_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('AlertInfo', alert_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            alert_vos, total_count = alert_service.list(params)
            return self.locator.get_info('AlertsInfo',
                                         alert_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('AlertService', metadata) as alert_service:
            return self.locator.get_info('StatisticsInfo', alert_service.stat(params))



import logging

from spaceone.core.service import *
from spaceone.monitoring.error import *
from spaceone.monitoring.model.project_alert_config_model import ProjectAlertConfig
from spaceone.monitoring.manager.identity_manager import IdentityManager
from spaceone.monitoring.manager.escalation_policy_manager import EscalationPolicyManager
from spaceone.monitoring.manager.project_alert_config_manager import ProjectAlertConfigManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ProjectAlertConfigService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_alert_config_mgr: ProjectAlertConfigManager = self.locator.get_manager('ProjectAlertConfigManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    def create(self, params):
        """Create project alert configuration

        Args:
            params (dict): {
                'project_id': 'str',
                'escalation_policy_id': 'str',
                'options': 'dict',
                'domain_id': 'str'
            }

        Returns:
            project_alert_config_vo (object)
        """

        project_id = params['project_id']
        escalation_policy_id = params.get('escalation_policy_id')
        domain_id = params['domain_id']

        identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
        escalation_policy_mgr: EscalationPolicyManager = self.locator.get_manager('EscalationPolicyManager')

        identity_mgr.get_project(project_id, domain_id)

        if escalation_policy_id:
            escalation_policy_vo = escalation_policy_mgr.get_escalation_policy(escalation_policy_id, domain_id)
            if escalation_policy_vo.scope == 'PROJECT' and escalation_policy_vo.project_id != project_id:
                raise ERROR_INVALID_ESCALATION_POLICY(escalation_policy_id=escalation_policy_id)

            params['escalation_policy'] = escalation_policy_vo

        else:
            escalation_policy_vo = escalation_policy_mgr.get_default_escalation_policy(domain_id)
            params['escalation_policy_id'] = escalation_policy_vo.escalation_policy_id
            params['escalation_policy'] = escalation_policy_vo

        return self.project_alert_config_mgr.create_project_alert_config(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    def update(self, params):
        """Update project alert configuration

        Args:
            params (dict): {
                'project_id': 'str',
                'escalation_policy_id': 'dict',
                'options': 'dict',
                'domain_id': 'str'
            }

        Returns:
            project_alert_config_vo (object)
        """
        
        project_id = params['project_id']
        escalation_policy_id = params.get('escalation_policy_id')
        domain_id = params['domain_id']

        project_alert_config_vo: ProjectAlertConfig = self.project_alert_config_mgr.get_project_alert_config(project_id,
                                                                                                             domain_id)

        if escalation_policy_id:
            escalation_policy_mgr: EscalationPolicyManager = self.locator.get_manager('EscalationPolicyManager')
            escalation_policy_vo = escalation_policy_mgr.get_escalation_policy(escalation_policy_id, domain_id)
            if escalation_policy_vo.scope == 'PROJECT' and escalation_policy_vo.project_id != project_id:
                raise ERROR_INVALID_ESCALATION_POLICY(escalation_policy_id=escalation_policy_id)

            params['escalation_policy'] = escalation_policy_vo

        if 'options' in params:
            if 'recovery_mode' not in params['options']:
                params['options']['recovery_mode'] = project_alert_config_vo.options.recovery_mode

            if 'notification_urgency' not in params['options']:
                params['options']['notification_urgency'] = project_alert_config_vo.options.notification_urgency

        return self.project_alert_config_mgr.update_project_alert_config_by_vo(params, project_alert_config_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    def delete(self, params):
        """Delete project alert configuration

        Args:
            params (dict): {
                'project_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.project_alert_config_mgr.delete_project_alert_config(params['project_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    @change_only_key({'escalation_policy_info': 'escalation_policy'})
    def get(self, params):
        """ Get project alert configuration

        Args:
            params (dict): {
                'project_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            project_alert_config_vo (object)
        """

        return self.project_alert_config_mgr.get_project_alert_config(params['project_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @change_only_key({'escalation_policy_info': 'escalation_policy'}, key_path='query.only')
    @append_query_filter(['project_id', 'escalation_policy_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['project_id'])
    def list(self, params):
        """ List project alert configurations

        Args:
            params (dict): {
                'project_id': 'str',
                'escalation_policy_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', // from meta
            }

        Returns:
            project_alert_config_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.project_alert_config_mgr.list_project_alert_configs(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['project_id'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.project_alert_config_mgr.stat_project_alert_configs(query)



import logging

from google.protobuf.json_format import MessageToDict

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_endpoint
from spaceone.core.error import *

__all__ = ['WebhookPluginConnector']

_LOGGER = logging.getLogger(__name__)


class WebhookPluginConnector(BaseConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = None

    def initialize(self, endpoint):
        static_endpoint = self.config.get('endpoint')

        if static_endpoint:
            endpoint = static_endpoint

        e = parse_endpoint(endpoint)
        self.client = pygrpc.client(endpoint=f'{e.get("hostname")}:{e.get("port")}', version='plugin')

    def init(self, options):
        response = self.client.Webhook.init({
            'options': options,
        }, metadata=self.transaction.get_connection_meta())

        return self._change_message(response)

    def verify(self, options):
        params = {
            'options': options
        }

        self.client.Webhook.verify(params, metadata=self.transaction.get_connection_meta())

    def parse_event(self, options, data):
        params = {
            'options': options,
            'data': data
        }

        response = self.client.Event.parse(params, metadata=self.transaction.get_connection_meta())
        return self._change_message(response)

    @staticmethod
    def _change_message(message):
        return MessageToDict(message, preserving_proto_field_name=True)



import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.monitoring.model.webhook_model import Webhook

_LOGGER = logging.getLogger(__name__)


class WebhookManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.webhook_model: Webhook = self.locator.get_model('Webhook')

    def create_webhook(self, params):
        def _rollback(webhook_vo):
            _LOGGER.info(f'[create_webhook._rollback] '
                         f'Delete webhook : {webhook_vo.name} '
                         f'({webhook_vo.webhook_id})')
            webhook_vo.delete()

        webhook_vo: Webhook = self.webhook_model.create(params)
        self.transaction.add_rollback(_rollback, webhook_vo)

        return webhook_vo

    def update_webhook(self, params):
        webhook_vo: Webhook = self.get_webhook(params['webhook_id'], params['domain_id'])
        return self.update_webhook_by_vo(params, webhook_vo)

    def update_webhook_by_vo(self, params, webhook_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_webhook_by_vo._rollback] Revert Data : '
                         f'{old_data["webhook_id"]}')
            webhook_vo.update(old_data)

        self.transaction.add_rollback(_rollback, webhook_vo.to_dict())

        updated_vo: Webhook = webhook_vo.update(params)

        cache.delete(f'webhook-data:{updated_vo.webhook_id}')

        return updated_vo

    def delete_webhook(self, webhook_id, domain_id):
        webhook_vo: Webhook = self.get_webhook(webhook_id, domain_id)

        cache.delete(f'webhook-data:{webhook_vo.webhook_id}')

        webhook_vo.delete()

    def get_webhook(self, webhook_id, domain_id, only=None):
        return self.webhook_model.get(webhook_id=webhook_id, domain_id=domain_id, only=only)

    def get_webhook_by_id(self, webhook_id):
        return self.webhook_model.get(webhook_id=webhook_id)

    def list_webhooks(self, query={}):
        return self.webhook_model.query(**query)

    def stat_webhooks(self, query):
        return self.webhook_model.stat(**query)



import logging
import datetime
from spaceone.core.service import *
from spaceone.core.utils import get_dict_value
from spaceone.monitoring.error import *
from spaceone.monitoring.manager.identity_manager import IdentityManager
from spaceone.monitoring.manager.inventory_manager import InventoryManager
from spaceone.monitoring.manager.secret_manager import SecretManager
from spaceone.monitoring.manager.data_source_manager import DataSourceManager
from spaceone.monitoring.manager.data_source_plugin_manager import DataSourcePluginManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class LogService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
        self.inventory_mgr: InventoryManager = self.locator.get_manager('InventoryManager')
        self.secret_mgr: SecretManager = self.locator.get_manager('SecretManager')
        self.data_source_mgr: DataSourceManager = self.locator.get_manager('DataSourceManager')
        self.ds_plugin_mgr: DataSourcePluginManager = self.locator.get_manager('DataSourcePluginManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'resource_id', 'start', 'domain_id'])
    def list(self, params):
        """ Get resource's logs

        Args:
            params (dict): {
                'data_source_id': 'str',
                'resource_id': 'str',
                'keyword': 'str',
                'start': 'datetime',
                'end': 'datetime',
                'sort': 'dict',
                'limit': 'int',
                'domain_id': 'str'
            }

        Returns:
            logs (list)
        """
        data_source_id = params['data_source_id']
        resource_id = params['resource_id']
        start = params['start']
        end = params.get('end', str(datetime.datetime.now()))
        domain_id = params['domain_id']
        secret_data = {}

        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)
        self._check_data_source_state(data_source_vo)
        self.plugin_initialize(data_source_vo)

        plugin_options = data_source_vo.plugin_info.options
        cloud_service_info = self.inventory_mgr.get_cloud_service(resource_id, domain_id)
        query = self.get_query_from_cloud_service(cloud_service_info, data_source_vo.plugin_info)
        secret = self.secret_mgr.get_secret_from_resource(cloud_service_info, data_source_vo, domain_id)
        secret_data = self.secret_mgr.get_secret_data(secret.get('secret_id', ''), domain_id)

        logs_info = self.ds_plugin_mgr.list_logs(secret.get('schema'), plugin_options, secret_data,
                                                 query, params.get('keyword'),
                                                 start, end, params.get('sort'), params.get('limit'))

        return {
            'results': logs_info['results'],
            'domain_id': domain_id
        }

    def plugin_initialize(self, data_source_vo):
        endpoint = self.ds_plugin_mgr.get_data_source_plugin_endpoint_by_vo(data_source_vo)
        self.ds_plugin_mgr.initialize(endpoint)

    @staticmethod
    def get_query_from_cloud_service(cloud_service_info, plugin_info):
        metadata = plugin_info.metadata
        required_keys = metadata.get('required_keys', [])

        if required_keys:
            query_key = required_keys[0]
            query = get_dict_value(cloud_service_info, query_key, default_value={})
            if query:
                return query
            else:
                raise ERROR_REQUIRED_KEYS_NOT_EXISTS(plugin_id=plugin_info.plugin_id)
        else:
            raise ERROR_REQUIRED_KEYS_NOT_EXISTS(plugin_id=plugin_info.plugin_id)

    @staticmethod
    def _check_data_source_state(data_source_vo):
        if data_source_vo.state == 'DISABLED':
            raise ERROR_DATA_SOURCE_STATE_DISABLED(data_source_id=data_source_vo.data_source_id)



import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.monitoring.error.project_alert_config import *
from spaceone.monitoring.model.project_alert_config_model import ProjectAlertConfig

_LOGGER = logging.getLogger(__name__)


class ProjectAlertConfigManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_alert_config_model: ProjectAlertConfig = self.locator.get_model('ProjectAlertConfig')

    def create_project_alert_config(self, params):
        def _rollback(project_alert_config_vo: ProjectAlertConfig):
            _LOGGER.info(f'[create_project_alert_config._rollback] '
                         f'Delete project alert config : {project_alert_config_vo.project_id}')
            project_alert_config_vo.delete()

        project_alert_config_vo: ProjectAlertConfig = self.project_alert_config_model.create(params)
        self.transaction.add_rollback(_rollback, project_alert_config_vo)

        return project_alert_config_vo

    def update_project_alert_config(self, params):
        project_alert_config_vo: ProjectAlertConfig = self.get_project_alert_config(params['project_id'],
                                                                                    params['domain_id'])
        return self.update_project_alert_config_by_vo(params, project_alert_config_vo)

    def update_project_alert_config_by_vo(self, params, project_alert_config_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_project_alert_config_by_vo._rollback] Revert Data : '
                         f'{old_data["project_id"]}')
            project_alert_config_vo.update(old_data)

        self.transaction.add_rollback(_rollback, project_alert_config_vo.to_dict())

        updated_vo: ProjectAlertConfig = project_alert_config_vo.update(params)

        cache.delete(f'project-alert-options:{updated_vo.domain_id}:{updated_vo.project_id}')
        cache.delete(f'escalation-policy-info:{updated_vo.domain_id}:{updated_vo.project_id}')
        cache.delete(f'auto-recovery:{updated_vo.domain_id}:{updated_vo.project_id}')

        return updated_vo

    def delete_project_alert_config(self, project_id, domain_id):
        project_alert_config_vo: ProjectAlertConfig = self.get_project_alert_config(project_id, domain_id)

        cache.delete(f'project-alert-options:{domain_id}:{project_id}')
        cache.delete(f'escalation-policy-info:{domain_id}:{project_id}')
        cache.delete(f'auto-recovery::{domain_id}:{project_id}')

        project_alert_config_vo.delete()

    def get_project_alert_config(self, project_id, domain_id, only=None):
        try:
            return self.project_alert_config_model.get(project_id=project_id, domain_id=domain_id, only=only)
        except ERROR_NOT_FOUND as e:
            raise ERROR_ALERT_FEATURE_IS_NOT_ACTIVATED(project_id=project_id)
        except Exception as e:
            raise e

    def list_project_alert_configs(self, query={}):
        return self.project_alert_config_model.query(**query)

    def stat_project_alert_configs(self, query):
        return self.project_alert_config_model.stat(**query)



import logging
import fnmatch

from spaceone.core.service import *
from spaceone.monitoring.error.event_rule import *
from spaceone.monitoring.model.event_rule_model import EventRule
from spaceone.monitoring.manager.identity_manager import IdentityManager
from spaceone.monitoring.manager.event_rule_manager import EventRuleManager

_LOGGER = logging.getLogger(__name__)

_SUPPORTED_CONDITION_KEYS = ['title', 'description', 'rule', 'resource_id', 'resource_name', 'resource_type',
                             'webhook_id', 'project_id', 'provider', 'account', 'additional_info.<key>']
_SUPPORTED_CONDITION_OPERATORS = ['eq', 'contain', 'not', 'not_contain']


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class EventRuleService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.event_rule_mgr: EventRuleManager = self.locator.get_manager('EventRuleManager')

    @transaction(append_meta={
        'authorization.scope': 'PROJECT',
        'authorization.require_project_id': True
    })
    @check_required(['conditions', 'conditions_policy', 'actions'])
    def create(self, params):
        """Create event rule

        Args:
            params (dict): {
                'name': 'str',
                'conditions': 'list',
                'conditions_policy': 'str',
                'actions': 'dict',
                'options': 'dict',
                'project_id': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            event_rule_vo (object)
        """

        project_id = params.get('project_id')
        domain_id = params['domain_id']

        identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')

        if project_id:
            identity_mgr.get_project(project_id, domain_id)
            params['scope'] = 'PROJECT'
        else:
            params['scope'] = 'GLOBAL'

        self._check_conditions(params['conditions'])
        self._check_actions(params['actions'])

        params['order'] = self._get_highest_order(params['scope'], params.get('project_id'), params['domain_id']) + 1

        return self.event_rule_mgr.create_event_rule(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['event_rule_id', 'domain_id'])
    def update(self, params):
        """Update event rule

        Args:
            params (dict): {
                'event_rule_id': 'dict',
                'name': 'str',
                'conditions': 'list',
                'conditions_policy': 'str',
                'actions': 'dict',
                'options': 'dict',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            event_rule_vo (object)
        """

        event_rule_id = params['event_rule_id']
        domain_id = params['domain_id']

        if 'conditions' in params:
            self._check_conditions(params['conditions'])

        if 'actions' in params:
            self._check_actions(params['actions'])

        event_rule_vo = self.event_rule_mgr.get_event_rule(event_rule_id, domain_id)
        return self.event_rule_mgr.update_event_rule_by_vo(params, event_rule_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['event_rule_id', 'order', 'domain_id'])
    def change_order(self, params):
        """ Get event rule

        Args:
            params (dict): {
                'event_rule_id': 'str',
                'order': 'int',
                'domain_id': 'str'
            }

        Returns:
            event_rule_vo (object)
        """

        event_rule_id = params['event_rule_id']
        order = params['order']
        domain_id = params['domain_id']

        self._check_order(order)

        target_event_rule_vo = self.event_rule_mgr.get_event_rule(event_rule_id, domain_id)

        if target_event_rule_vo.order == order:
            return target_event_rule_vo

        highest_order = self._get_highest_order(target_event_rule_vo.scope, target_event_rule_vo.project_id,
                                                target_event_rule_vo.domain_id)

        if order > highest_order:
            raise ERROR_INVALID_PARAMETER(key='order',
                                          reason=f'There is no event rules greater than the {str(order)} order.')

        event_rule_vos = self._get_all_event_rules(target_event_rule_vo.scope, target_event_rule_vo.project_id,
                                                   domain_id, target_event_rule_vo.event_rule_id)
        event_rule_vos.insert(order-1, target_event_rule_vo)

        i = 0
        for event_rule_vo in event_rule_vos:
            if target_event_rule_vo != event_rule_vo:
                self.event_rule_mgr.update_event_rule_by_vo({'order': i+1}, event_rule_vo)

            i += 1

        return self.event_rule_mgr.update_event_rule_by_vo({'order': order}, target_event_rule_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['event_rule_id', 'domain_id'])
    def delete(self, params):
        """Delete event rule

        Args:
            params (dict): {
                'event_rule_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        event_rule_id = params['event_rule_id']
        domain_id = params['domain_id']

        event_rule_vo: EventRule = self.event_rule_mgr.get_event_rule(event_rule_id, domain_id)

        scope = event_rule_vo.scope
        project_id = event_rule_vo.project_id

        self.event_rule_mgr.delete_event_rule_by_vo(event_rule_vo)

        event_rule_vos = self._get_all_event_rules(scope, project_id, domain_id)

        i = 0
        for event_rule_vo in event_rule_vos:
            self.event_rule_mgr.update_event_rule_by_vo({'order': i+1}, event_rule_vo)
            i += 1

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['event_rule_id', 'domain_id'])
    def get(self, params):
        """ Get event rule

        Args:
            params (dict): {
                'event_rule_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            event_rule_vo (object)
        """

        return self.event_rule_mgr.get_event_rule(params['event_rule_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['event_rule_id', 'name', 'scope', 'project_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['event_rule_id', 'name'])
    def list(self, params):
        """ List escalation polices

        Args:
            params (dict): {
                'event_rule_id': 'str',
                'name': 'str',
                'scope': 'str',
                'project_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', // from meta
            }

        Returns:
            event_rule_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.event_rule_mgr.list_event_rules(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['event_rule_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.event_rule_mgr.stat_event_rules(query)

    @staticmethod
    def _check_conditions(conditions):
        for condition in conditions:
            key = condition.get('key')
            value = condition.get('value')
            operator = condition.get('operator')

            if not (key and value and operator):
                raise ERROR_INVALID_PARAMETER(key='conditions', reason='Condition should have key, value and operator.')

            if key not in _SUPPORTED_CONDITION_KEYS and not fnmatch.fnmatch(key, 'additional_info.*'):
                raise ERROR_INVALID_PARAMETER(key='conditions.key',
                                              reason=f'Unsupported key. '
                                                     f'({" | ".join(_SUPPORTED_CONDITION_KEYS)})')
            if operator not in _SUPPORTED_CONDITION_OPERATORS:
                raise ERROR_INVALID_PARAMETER(key='conditions.operator',
                                              reason=f'Unsupported operator. '
                                                     f'({" | ".join(_SUPPORTED_CONDITION_OPERATORS)})')

    @staticmethod
    def _check_actions(actions):
        if 'change_assignee' in actions:
            # Check User
            pass

        if 'change_urgency' in actions:
            if actions['change_urgency'] not in ['HIGH', 'LOW']:
                raise ERROR_INVALID_PARAMETER(key='actions.change_urgency',
                                              reason=f'Unsupported urgency. (HIGH | LOW)')

        if 'change_project' in actions:
            # Check Project
            pass

        for project_id in actions.get('add_project_dependency', []):
            # Check Project
            pass

        for responder in actions.get('add_responder', []):
            # Check User
            pass

    @staticmethod
    def _check_order(order):
        if order <= 0:
            raise ERROR_INVALID_PARAMETER(key='order', reason='The order must be greater than 0.')

    def _get_highest_order(self, scope, project_id, domain_id):
        query = {
            'filter': [
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                },
                {
                    'k': 'scope',
                    'v': scope,
                    'o': 'eq'
                },
                {
                    'k': 'project_id',
                    'v': project_id,
                    'o': 'eq'
                }
            ],
            'count_only': True
        }

        event_rule_vos, total_count = self.event_rule_mgr.list_event_rules(query)
        return total_count

    def _get_all_event_rules(self, scope, project_id, domain_id, exclude_event_rule_id=None):

        query = {
            'filter': [
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                },
                {
                    'k': 'scope',
                    'v': scope,
                    'o': 'eq'
                },
                {
                    'k': 'project_id',
                    'v': project_id,
                    'o': 'eq'
                },
            ],
            'sort': {
                'key': 'order'
            }
        }
        if exclude_event_rule_id is not None:
            query['filter'].append({
                'k': 'event_rule_id',
                'v': exclude_event_rule_id,
                'o': 'not'
            })

        event_rule_vos, total_count = self.event_rule_mgr.list_event_rules(query)
        return list(event_rule_vos)



from spaceone.api.monitoring.v1 import escalation_policy_pb2, escalation_policy_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class EscalationPolicy(BaseAPI, escalation_policy_pb2_grpc.EscalationPolicyServicer):

    pb2 = escalation_policy_pb2
    pb2_grpc = escalation_policy_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            return self.locator.get_info('EscalationPolicyInfo', escalation_policy_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            return self.locator.get_info('EscalationPolicyInfo', escalation_policy_service.update(params))

    def set_default(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            return self.locator.get_info('EscalationPolicyInfo', escalation_policy_service.set_default(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            escalation_policy_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            return self.locator.get_info('EscalationPolicyInfo', escalation_policy_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            escalation_policy_vos, total_count = escalation_policy_service.list(params)
            return self.locator.get_info('EscalationPoliciesInfo',
                                         escalation_policy_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('EscalationPolicyService', metadata) as escalation_policy_service:
            return self.locator.get_info('StatisticsInfo', escalation_policy_service.stat(params))



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class Error(EmbeddedDocument):
    error_code = StringField(max_length=128)
    message = StringField(max_length=2048)


class Job(MongoModel):
    job_id = StringField(max_length=40, generate_id='job', unique=True)
    domain_id = StringField(max_length=40, required=True)
    status = StringField(max_length=20, default='IN_PROGRESS', choices=('IN_PROGRESS', 'SUCCESS', 'ERROR', 'TIMEOUT'))
    errors = ListField(EmbeddedDocumentField(Error, default=None, null=True))
    total_tasks = IntField(default=0)
    remained_tasks = IntField(default=0)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)
    finished_at = DateTimeField(default=None, null=True)

    meta = {
        'updatable_fields': [
            'status',
            'errors',
            'total_tasks',
            'remained_tasks',
            'finished_at'
        ],
        'ordering': [
            '-created_at'
        ],
        'indexes': [
            # 'job_id',
            'domain_id',
            'status',
            'total_tasks',
            'remained_tasks',
            'created_at'
        ]
    }
